/*******************************************************************************
 *                                                                             *
 *        Program : Inclusive Prefix Sum for vector                            *
 *                                                                             *
 *        Author  : Kim, Dongmin (kimd@top.cis.syr.edu)                        *
 *                                                                             *
 *        This program simulates the prefix sum algorithm for vector.          *
 *                                                                             *
 *        The vector size should be multiple of total processor number.        *
 *                                                                             *
 *******************************************************************************/


#include <cm/cmmd.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define Max 1024


/*******************************************************************************
 *                                                                             *
 *   Global variables                                                          *
 *                                                                             *
 *******************************************************************************/
int    my_id;  /* The ID of the Processor */
int    nproc;  /* Total # of Processors */
int    v_size; /* Vector Size */
int    digit;  /* binary digit */
int    in_mesg[Max]; /* received data */
int    Vector[Max];  /* original data */
int    Prefix[Max];  /* result data */
int    Right[Max][Max];   /* Sums from right child process by step.
				  1st dimension is for step, and
				  2nd dimension is for the sums
				  from the right child process.          */
int    Sum[Max];     /* Sum from parent process */

int    partner;    /* The processor ID with which the data is exchanged */
int    right;      /* The processor ID of right child */
int    parent;     /* The processor ID of parent */



/*******************************************************************************
 *   Procedure :                                                               *
 *   Input     :                                                               *
 *   Output    :                                                               *
 *   Function  :                                                               *
 *******************************************************************************/
/*******************************************************************************
 *   Procedure :                                                               *
 *   Input     :                                                               *
 *   Output    :                                                               *
 *   Function  :                                                               *
 *******************************************************************************/
show(id, start, size)
     int id, start, size;
{
  int pn;
  int i;


  CMMD_fset_io_mode(stdout, CMMD_independent);
  for ( pn = 0 ; pn < nproc ; pn++ ) {
    if ( id == pn ) {
      printf("***********************************************\n");
      printf("Processor %2d => \n", id);
      for ( i = start ; i < start+size ; i++ )
	printf("%d  :  %3d\n", i, Vector[i]);

      printf("***********************************************\n");
      fflush(stdout);
    }

    CMMD_sync_with_nodes();
  }
}

show2(id, partner, start, size, check)
     int id, partner, start, size, check;
{
  int pn;
  int i;


  CMMD_fset_io_mode(stdout, CMMD_independent);
  for ( pn = 0 ; pn < nproc ; pn++ ) {
    if ( id == pn ) {
      printf("***********************************************\n");
      printf("Processor %2d   Partner %2d   check %d\n", id, partner, check);
      for ( i = start ; i < start+size ; i++ )
	printf("%d  :  %3d\n", i, in_mesg[i]);

      printf("***********************************************\n");
      fflush(stdout);
    }

    CMMD_sync_with_nodes();
  }
}

show3(id, digit, size)
     int id, digit, size;
{
  int pn ;
  int i, j, k, temp_size;


  CMMD_fset_io_mode(stdout, CMMD_independent);

  temp_size = size;
  for ( pn = 0 ; pn < nproc ; pn++ ) {
    if ( id == pn ) {
      printf("***********************************************\n");
      printf("Processor %2d   \n", id);
      for ( i = 0 ; i < digit ; i++ ) {
	for ( j = 0 ; j < v_size ; j++ ) {
	  printf("Right[%d][%d]  :  %3d\n", i, j, Right[i][j]);
	  temp_size = temp_size / 2;
	}
      }

      printf("***********************************************\n");
      fflush(stdout);
    }

    CMMD_sync_with_nodes();
  }
}

show4(id, ch, value)
     char* ch;
     int value, id;
{

  int    pn;

  CMMD_fset_io_mode(stdout, CMMD_independent);

  for ( pn = 0 ; pn < nproc ; pn++ ) {
    if ( id == pn ) {
      printf("***********************************************\n");
      printf("Processor %2d   \n", id);
	  printf("%s : %d\n", ch, value);

      printf("***********************************************\n");
      fflush(stdout);
    }

    CMMD_sync_with_nodes();
  }
}






/*******************************************************************************
 *   Procedure : add_n_store                                                   *
 *   Input     : int start_index, int data_size, int step                      *
 *   Output    : None                                                          *
 *   Function  : Add the two arrays from the start_index by data_size, and     *
                 store the right child's data.                                 *
 *******************************************************************************/add_n_store(start_index, data_size, step)
     int start_index,
         data_size,
         step;
{
  int i;

  if ( right == partner ) {
    for ( i = start_index ; i < start_index + data_size ; i++ ) {
      Vector[i] += in_mesg[i];
      Right[step][i] = in_mesg[i];
    }
  }
  else {
    for ( i = start_index ; i < start_index + data_size ; i++ ) {
      Right[step][i] = Vector[i];
      Vector[i] += in_mesg[i];
    }
  }
}

/*******************************************************************************
 *   Procedure : Ilog2()                                                       *
 *   Input     : int n                                                         *
 *   Output    : log2(n)                                                       *
 *   Function  : return the integer value of log2(n) if it is integer.         *
 *               If not, return -1.                                            *
 *******************************************************************************/
int Ilog2(n)
     int n;
{
  int i = 0;
  int n1 = n;

  if ( n < 1 ) return -1;

  while ( n1 >>= 1 )
    ++i;

  if ( 1 << i != n )
    return -1;
  else
    return i;
}

/*******************************************************************************
 *   Procedure : pow2()                                                        *
 *   Input     : int x                                                         *
 *   Output    : 2^x                                                           *
 *   Function  : power of 2 for the input value                                *
 *******************************************************************************/
int pow2(x)
     int x;
{
  int p = 1;
  int i;

  for ( i = 0 ; i < x ; i++ )
    p <<= 1;

  return p;
}

/*******************************************************************************
 *                                                                             * 
 *                              MAIN  PROGRAM                                  *
 *                                                                             *
 *******************************************************************************/
  
main(argc, argv)
     int argc;
     char *argv[];
{
  FILE   *fd; /* input vector file name */

  int    i, j, k;
  int    local_size; /* Local vector size */
  int    step;
  int    size;       /* size of integer */
  int    data_size;  /* data size */
  int    digit;      /* total digit in binary */
  int    check;      /* check if the i-th bit is 1 */
  int    start_send; /* the start index to be sent */
  int    start_recv; /* the start index to be calculated */
  int    start;

  /*-------------------------  Initialize  -----------------------------*/

  CMMD_disable_safety();
  CMMD_fset_io_mode(stdin, CMMD_sync_bc);
  CMMD_fset_io_mode(stdout, CMMD_sync_bc);
  CMMD_fset_io_mode(stderr, CMMD_independent);

  if (argc < 3) {
    printf("Usage: (1) input vector file name \n");
    printf("Usage: (2) the number of processor(s) ( 2 - 32 ) \n");
    printf("Usage: (3) vector size \n");
    exit(1);
  }

/*  buf_size = Nlocal * Nlocal * sizeof(double);*/

/*  ori_nproc = CMMD_partition_size(); */

  my_id = CMMD_self_address(); 

  nproc = atoi(argv[2]);
  v_size = atoi(argv[3]);

  local_size = v_size / nproc;

  i = CMMD_reset_partition_size(nproc);

  if (i != nproc) 
    CMMD_error("\nERROR: CMMD Partition failed!\n");

  /* all processors read the vector from the file */

  fd = fopen(argv[1], "r");
  if ( fd == NULL ) {
    printf("\ncannot find file %s!\n", argv[1]);
    exit(1);
  }

  for ( i = 0 ; i < my_id * v_size ; i++ ) {
      fscanf(fd, "%d", &Vector[0]);
  }

  for ( i = 0 ; i < v_size ; i++ ) {
      fscanf(fd, "%d", &Vector[i]);
  }

  fclose(fd);
/*
  for ( i = 0 ; i < v_size ; i++ )
    Vector[i] = Vector[i];
*/
  show(my_id, 0, v_size);

  size = sizeof(int);
  data_size = v_size;

  start_recv = 0;

  CMMD_sync_with_nodes();

  digit = Ilog2(nproc);

/* Accumulation */
  for ( step = 0 ; step < digit ; step++ ) {

    data_size = data_size / 2;

    start_send = start_recv;
    
    check = my_id & pow2(step);

    if ( check == 0 )
      start_send += data_size;
    else
      start_recv += data_size;

    partner = my_id ^ pow2(step);

    right = my_id | pow2(step);

    CMMD_swap( partner,
	       &in_mesg[start_recv], data_size * size,  /* receive */
	       &Vector[start_send],  data_size * size); /* send */

/*    show2(my_id, partner, start_recv, data_size, check);*/
/*    show2(my_id, partner, start_recv, data_size, right);*/

    add_n_store(start_recv, data_size, step);


  }

  show(my_id, 0, v_size);
/*
  show(my_id, start_recv, local_size);
*/
  show3(my_id, digit, v_size / 2);

/*  data_size = Ilog2(v_size / 2);*/

  show4(my_id, "data_size", data_size);
  start_send = start_recv;

/* Distribution */

/*  show4(my_id, "start_send", start_send);*/

  for ( step = digit - 1 ; step >= 0 ; step-- ) {

    check = my_id & pow2(step);

    if ( check == 0 )
      start_recv += data_size;
    else
      start_recv -= data_size;

    partner = my_id ^ pow2(step);

    right = my_id | pow2(step);

    if ( partner == right ) {
      CMMD_swap( partner,
	         &Vector[start_recv], data_size * size,  /* receive */
	         &Vector[start_send], data_size * size); /* send */
      for ( i = start_send ; i < start_send + data_size ; i++ )
	Vector[i] -= Right[step][i];

    }
    else {
      for ( i = start_send ; i < start_send + data_size ; i++ )
	in_mesg[i] = Vector[i] - Right[step][i];
      CMMD_swap( partner,
	         &Vector[start_recv],  data_size * size,  /* receive */
	         &in_mesg[start_send], data_size * size); /* send */

    }


/*
    show2(my_id, partner, start_recv, data_size, check);
    show(my_id, start_recv, data_size);
*/
/*show4(my_id, "data_size", data_size);*/
    data_size = data_size * 2;

    if ( start_send > start_recv )
      start_send = start_recv;
    start_recv = start_send;
  }

  show(my_id, 0, v_size);

}

